/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include "util.h"
#include "configfile.h"
#include "timetracker.h"
#include "kis_netclient.h"

// Reconnect trigger
int KisCliReconEvent(Timetracker::timer_event *evt, void *parm,
					 GlobalRegistry *globalreg) {
	KismetClient *alc = (KismetClient *) parm;

	alc->Reconnect();

	return 1;
}

KismetClient::KismetClient() {
    fprintf(stderr, "*** KismetClient called with no global registry reference\n");
    globalreg = NULL;
    tcpcli = NULL;
	filters = NULL;
}

KismetClient::KismetClient(GlobalRegistry *in_globalreg) : 
	ClientFramework(in_globalreg) {
    // The only connection type we support right now is TCP so just make it manually
    // this could be parsed from the config file when we support other transport
    // layers
    tcpcli = new TcpClient(globalreg);

    // Attach it to ourselves and opposite
    RegisterNetworkClient(tcpcli);
    tcpcli->RegisterClientFramework(this);

    // Initialize all our stuff
    cmdid = 1;

    reconnect_attempt = -1;

    server_identified = 0;
    
    maxalerts = 500;

	n_networks = n_packets = n_crypt = n_weak = packetrate = 0;

    // Fill in the protocol default field lists
    protocol_default_map["ALERT"] = "sec,usec,bssid,source,dest,other,channel,"
		"header,text";
    protocol_default_map["KISMET"] = "version,starttime,servername,timestamp,"
        "channelhop,newversion";
    protocol_default_map["NETWORK"] = "bssid,type,ssid,beaconinfo,llcpackets,"
		"datapackets,cryptpackets,weakpackets,channel,wep,firsttime,lasttime,"
		"atype,rangeip,gpsfixed,minlat,minlon,minalt,minspd,"
		"maxlat,maxlon,maxalt,maxspd,bestlat,bestlon,bestalt,"
		"agglat,agglon,aggalt,aggpoints,quality,signal,noise,bestquality,"
		"bestsignal,bestnoise,cloaked,decrypted,beaconrate,maxrate,carrierset,"
		"maxseenrate,encodingset,datasize";
	protocol_default_map["REMOVE"] = "bssid";
	protocol_default_map["ERROR"] = "text";
	protocol_default_map["STATUS"] = "text";
	protocol_default_map["INFO"] = "networks,packets,crypt,weak,rate";

	protocol_map["KISMET"] = 1;
	protocol_map["ALERT"] = 1;
	protocol_map["NETWORK"] = 1;
	protocol_map["REMOVE"] = 1;
	protocol_map["STATUS"] = 1;
	protocol_map["ERROR"] = 1;
	protocol_map["INFO"] = 1;

    // Parse the config file and enable the tcpclient
    char temphost[128];
    if (sscanf(globalreg->kismetui_config->FetchOpt("host").c_str(), "%128[^:]:%d", 
               temphost, &port) != 2) {
        globalreg->messagebus->InjectMessage("Invalid Kismet host in config, "
											 "host:port required", MSGFLAG_FATAL);
        globalreg->fatal_condition = 1;
        return;
    }
    snprintf(host, MAXHOSTNAMELEN, "%s", temphost);

	if (globalreg->kismetui_config->FetchOpt("maxalerts") != "") {
		if ((sscanf(globalreg->kismetui_config->FetchOpt("maxalerts").c_str(), "%d",
					&maxalerts)) != 1) {
			globalreg->messagebus->InjectMessage("Invalid 'maxalerts' value in "
												 "config file.", MSGFLAG_FATAL);
			globalreg->fatal_condition = 1;
			return;
		}
	}

	if (globalreg->kismetui_config->FetchOpt("maxstatus") != "") {
		if ((sscanf(globalreg->kismetui_config->FetchOpt("maxstatus").c_str(), "%d",
					&maxstatus)) != 1) {
			globalreg->messagebus->InjectMessage("Invalid 'maxstatus' value in "
												 "config file.", MSGFLAG_FATAL);
			globalreg->fatal_condition = 1;
			return;
		}
	}

    if (globalreg->kismetui_config->FetchOpt("reconnect") == "true") {
        globalreg->messagebus->InjectMessage("Enabling reconnection to the Kismet "
											 "server if the link is lost", 
											 MSGFLAG_INFO);
        reconnect_attempt = 0;
    }

	if (globalreg->kismetui_config->FetchOpt("filter") != "") {
		filters = new filter_set;
		if (ParseFilterConf(globalreg->kismetui_config->FetchOpt("filter"),
							filters, globalreg) < 0) {
			globalreg->messagebus->InjectMessage("Could not process filters "
												 "from the config file",
												 MSGFLAG_FATAL);
			globalreg->fatal_condition = 1;
			delete filters;
			filters = NULL;
			return;
		}
	}

    if (tcpcli->Connect(host, port) < 0) {
        globalreg->messagebus->InjectMessage("Could not create initial connection to "
                                             "the Kismet server", MSGFLAG_ERROR);

		if (reconnect_attempt < 0) {
			globalreg->messagebus->InjectMessage("Connection failed and reconnecting "
												 "disabled.", MSGFLAG_FATAL);
			globalreg->fatal_condition = 1;

			return;
		}
		
        last_disconnect = time(0);
    } else {
        // Turn on the protocols in our map, if we have any.  This is for reconnecting
        // to a server and preserving the requested protocols
        for (map<string, int>::iterator prot = protocol_map.begin();
             prot != protocol_map.end(); ++prot) {
            EnableProtocol(prot->first);

            if (globalreg->fatal_condition)
                return;
        }
    }

	if (reconnect_attempt >= 0)
		reconid = 
			globalreg->timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * 5,
												  NULL, 1, &KisCliReconEvent,
												  (void *) this);
	else
		reconid = -1;

    snprintf(errstr, STATUS_MAX, "Using Kismet server on %s:%d", host, port);
    globalreg->messagebus->InjectMessage(errstr, MSGFLAG_INFO);
}

KismetClient::~KismetClient() {
    if (tcpcli != NULL) {
        tcpcli->KillConnection();
        delete tcpcli;
    }
    if (reconid >= 0 && globalreg != NULL)
        globalreg->timetracker->RemoveTimer(reconid);
}

int KismetClient::KillConnection() {
    if (tcpcli != NULL)
        tcpcli->KillConnection();

    return 1;
}

int KismetClient::Shutdown() {
    if (tcpcli != NULL) {
        tcpcli->FlushRings();
        tcpcli->KillConnection();
    }

    return 1;
}

int KismetClient::Valid() {
	if (tcpcli == NULL)
		return 0;

	return tcpcli->Valid();
}

void KismetClient::EnableProtocol(string in_protocol) {
    string fields;
    char data[1024];

    if (protocol_default_map.find(in_protocol) != protocol_default_map.end())
        fields = protocol_default_map[in_protocol];
    else
        fields = "*";

    // We don't care about ACK or errors, because we suck
    snprintf(data, 1024, "ENABLE %s %s\n", in_protocol.c_str(), fields.c_str());

    protocol_map[in_protocol] = 1;
    
	InjectCommand(data);
}

int KismetClient::InjectCommand(string in_cmd) {
    if (tcpcli->Valid() == 0) {
		return 0;
    }

    int curid = cmdid++;
    char cmdheader[32];

    snprintf(cmdheader, 32, "!%d ", curid);
    string fullcmd = string(cmdheader) + in_cmd;

    if (tcpcli->Valid() && tcpcli->WriteData((void *) fullcmd.c_str(), 
											 fullcmd.length()) < 0 ||
        globalreg->fatal_condition) {
        last_disconnect = time(0);
        return -1;
    }

    return curid;
}

int KismetClient::Reconnect() {
	if (tcpcli->Valid())
		return 1;
	
    if (tcpcli->Connect(host, port) < 0) {
        globalreg->messagebus->InjectMessage("Could not reconnect to Kismet server", 
											 MSGFLAG_ERROR);
        reconnect_attempt++;
        last_disconnect = time(0);
        return 0;
    } else {
        // Turn on the protocols in our map, if we have any.  This is for reconnecting
        // to a server and preserving the requested protocols
        for (map<string, int>::iterator prot = protocol_map.begin();
             prot != protocol_map.end(); ++prot) {
            EnableProtocol(prot->first);

            if (globalreg->fatal_condition)
                return 0;
        }
    }

	globalreg->messagebus->InjectMessage("Reconnected to the Kismet server",
										 MSGFLAG_INFO);
    
    return 1;
}

int KismetClient::ParseData() {
    int len, rlen;
    char *buf;
    string strbuf;

    // Scratch variables for parsing data
    char header[65];
    char bssid_str[18];
    mac_addr bssid;

    len = netclient->FetchReadLen();
    buf = new char[len + 1];
    
    if (netclient->ReadData(buf, len, &rlen) < 0) {
        globalreg->messagebus->InjectMessage("KismetClient::ParseData failed to "
											 "fetch data from the tcp connection.", 
											 MSGFLAG_ERROR);
        return -1;
    }
    buf[len] = '\0';

    // Parse without including partials, so we don't get a fragmented command 
    // out of the buffer
    vector<string> inptok = StrTokenize(buf, "\n", 0);
    delete[] buf;

    // Bail on no useful data
    if (inptok.size() < 1) {
        return 0;
    }

    for (unsigned int it = 0; it < inptok.size(); it++) {
        // No matter what we've dealt with this data block
        netclient->MarkRead(inptok[it].length() + 1);

        // Pull the header out to save time -- cheaper to parse the header and 
		// then the data than to try to parse an entire data string just to find 
		// out what protocol we are
        // 
        // Protocol parsers should be dynamic so that we can have plugins in the 
		// framework able to handle a proto, but right now thats a hassle

        if (sscanf(inptok[it].c_str(), "*%64[^:]", header) < 1) {
            continue;
        }

        // Nuke the header off the string
        inptok[it].erase(0, (size_t) strlen(header) + 3);

        if (!strncmp(header, "TERMINATE", 64)) {
            snprintf(errstr, STATUS_MAX, "Kismet server terminated.");
            globalreg->messagebus->InjectMessage(errstr, MSGFLAG_ERROR);
           
            netclient->KillConnection();
            
            continue;
        } else if (!strncmp(header, "KISMET", 64)) {
            // Parse kismet protocol or skip
			// We don't really care much whats in here
    		int junkmajor, junkminor, junktiny;
			int start_time;
			char servername[33];
			char build[25];
			int channel_hop;
			char major[25], minor[25], tiny[25];

            if (sscanf(inptok[it].c_str(), "%d.%d.%d %d \001%32[^\001]\001 %24s %d "
                       "%24[^.].%24[^.].%24s",
                       &junkmajor, &junkminor, &junktiny, 
                       (int *) &start_time, servername, 
                       build, &channel_hop,
                       major, minor, tiny) < 7)
                continue;

            server_identified = 1;

			remote_version = string(major) + string(minor) + string(tiny);
			remote_starttime = start_time;
			remote_servername = string(servername);

		} else if (!strncmp(header, "INFO", 64)) {
			sscanf(inptok[it].c_str(), "%d %d %d %d %d", &n_networks, 
				   &n_packets, &n_crypt, &n_weak, &packetrate);
		} else if (!strncmp(header, "ALERT", 64)) {
			// Parse the ALERT protocol
			char a_bssid_str[18];
			char a_src_str[18];
			char a_dest_str[18];
			char a_other_str[18];
			int a_ch;
			char a_hdr[128];
			char a_txt[2048];
			long int a_tvsec, a_tvusec;

			if (sscanf(inptok[it].c_str(), "%ld %ld %17s %17s %17s %17s %d %129s "
					   "\001%2047[^\001]\001",
					   &a_tvsec, &a_tvusec, a_bssid_str, a_src_str,
					   a_dest_str, a_other_str, &a_ch, a_hdr, a_txt) < 9)
				continue;

			alert_info *newal = new alert_info;

			newal->alert_ts.tv_sec = a_tvsec;
			newal->alert_ts.tv_usec = a_tvusec;
			newal->bssid = a_bssid_str;
			newal->source = a_src_str;
			newal->dest = a_dest_str;
			newal->other = a_other_str;
			newal->channel = a_ch;
			newal->header = string(a_hdr);
			newal->text = string(a_txt);

			if (alerts.size() > maxalerts) {
				delete alerts[0];
				alerts.erase(alerts.begin());
			}
			alerts.push_back(newal);
		} else if (!strncmp(header, "STATUS", 64)) {
			if (status_vec.size() > maxstatus) {
				status_vec.erase(status_vec.begin());
			}

			status_vec.push_back(inptok[it]);

		} else if (!strncmp(header, "ERROR", 64)) {
			if (status_vec.size() > maxstatus) {
				status_vec.erase(status_vec.begin());
			}

			status_vec.push_back(string("ERROR: ") + inptok[it]);

		} else if (!strncmp(header, "NETWORK", 64)) {
			// Smarter tokenization to handle quoted field buffers
			vector<string> net_toks = SmarterStrTokenize(inptok[it], " ", 1);

			mac_addr bssid = mac_addr(net_toks[0].c_str());

			if (filters != NULL) {
				printf("Filterting....\n");
				int fmatch = (filters->bssid_filter.find(bssid) !=
							  filters->bssid_filter.end());

				if ((fmatch && filters->bssid_invert) ||
					(!fmatch && !filters->bssid_invert)) {
					printf("Filtered: %s\n", net_toks[0].c_str());
					continue;
				}
			}

			network_info *netinf;
			map<mac_addr, network_info *>::const_iterator nitr;

			if ((nitr = network_map.find(bssid)) != network_map.end()) {
				netinf = nitr->second;
			} else {
				netinf = new network_info;
			}

			network_map[bssid] = netinf;

			netinf->bssid = bssid;
			
			if (sscanf(net_toks[1].c_str(), "%d", 
					   (int *) &(netinf->type)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			// This will break if we ever have a more complex client
			// attached to us
			if (netinf->type == (wireless_network_type) network_remove) {
				network_map.erase(bssid);
				delete netinf;
				continue;
			}

			netinf->ssid = net_toks[2];
			netinf->beaconinfo = net_toks[3];

			if (sscanf(net_toks[4].c_str(), "%d", 
					   (int *) &(netinf->llcpackets)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}
			if (sscanf(net_toks[5].c_str(), "%d", 
					   (int *) &(netinf->datapackets)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}
			if (sscanf(net_toks[6].c_str(), "%d", 
					   (int *) &(netinf->cryptpackets)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}
			if (sscanf(net_toks[7].c_str(), "%d", 
					   (int *) &(netinf->weakpackets)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}
			if (sscanf(net_toks[8].c_str(), "%d", 
					   (int *) &(netinf->channel)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}
			if (sscanf(net_toks[9].c_str(), "%d", 
					   (int *) &(netinf->wep)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[10].c_str(), "%d", 
					   (int *) &(netinf->firsttime)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[11].c_str(), "%d", 
					   (int *) &(netinf->lasttime)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[12].c_str(), "%d", 
					   (int *) &(netinf->atype)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			netinf->rangeip = net_toks[13];
			netinf->gpsfixed = net_toks[14];
			netinf->minlat = net_toks[15];
			netinf->minlon = net_toks[16];
			netinf->minalt = net_toks[17];
			netinf->minspd = net_toks[18];
			netinf->maxlat = net_toks[19];
			netinf->maxlon = net_toks[20];
			netinf->maxalt = net_toks[21];
			netinf->maxspd = net_toks[22];
			netinf->bestlat = net_toks[23];
			netinf->bestlon = net_toks[24];
			netinf->bestalt = net_toks[25];
			netinf->agglat = net_toks[26];
			netinf->agglon = net_toks[27];
			netinf->aggalt = net_toks[28];
			netinf->aggpoints = net_toks[29];
			netinf->quality = net_toks[30];
			netinf->signal = net_toks[31];
			netinf->noise = net_toks[32];
			netinf->bestquality = net_toks[33];
			netinf->bestsignal = net_toks[34];
			netinf->bestnoise = net_toks[35];

			if (sscanf(net_toks[36].c_str(), "%d", 
					   (int *) &(netinf->cloaked)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}
			if (sscanf(net_toks[27].c_str(), "%d", 
					   (int *) &(netinf->decrypted)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[38].c_str(), "%d", 
					   (int *) &(netinf->beaconrate)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[39].c_str(), "%f", 
					   (float *) &(netinf->maxrate)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[40].c_str(), "%d", 
					   (int *) &(netinf->carrierset)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[41].c_str(), "%d", 
					   (int *) &(netinf->maxseenrate)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[42].c_str(), "%d", 
					   (int *) &(netinf->encodingset)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

			if (sscanf(net_toks[43].c_str(), "%ld", 
					   (long int *) &(netinf->datasize)) != 1) {
				globalreg->messagebus->InjectMessage("Invalid NETWORK protocol",
													 MSGFLAG_ERROR);
			}

		} else if (!strncmp(header, "REMOVE", 64)) {
			// Smarter tokenization to handle quoted field buffers
			vector<string> net_toks = SmarterStrTokenize(inptok[it], " ", 1);

			mac_addr bssid = mac_addr(net_toks[0].c_str());

			network_info *netinf;
			map<mac_addr, network_info *>::const_iterator nitr;

			if ((nitr = network_map.find(bssid)) != network_map.end()) {
				netinf = nitr->second;
			} else {
				continue;
			}

			network_map.erase(bssid);
			delete netinf;
		}

    }

    return 1;
}

